/* tcpcat.c
 *
 * Copyright (C) 1996 by Mike Gleason.
 * All Rights Reserved.
 *
 * This may be used and distributed freely.
 *
 */

#ifdef HAVE_CONFIG_H
#	include "config.h"
#else
#	define HAVE_STRERROR 1
#endif

#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <errno.h>
#include <fcntl.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <signal.h>
#include <setjmp.h>
#include <time.h>
#include <ctype.h>

#ifndef INADDR_NONE
#	define INADDR_NONE ((unsigned long)(-1L))
#endif

struct sockaddr_in gServerAddr;
char *gProgName;
jmp_buf gJmp;
unsigned int gReadTimeout = 10;
int gPrintableOnly = 0;
int gPause = 0;
int gOriginatingPort = 0;
int gGotSig = 0;

extern int optind;
extern char *optarg;

static void
Perror(const char *const err)
{
	if (errno <= 0)
		fprintf(stderr, "%s :: %s.\n", gProgName, err);
	else {
#ifdef HAVE_STRERROR
		fprintf(stderr, "%s :: %s : %s.\n",
			gProgName, err, strerror(errno));
#else
		fprintf(stderr, "%s :: %s (errno=%d).\n",
			gProgName, err, errno);
#endif
	}
}	/* Perror */




static void
SigUsr1(int sigNum)
{
	gGotSig = sigNum;
	return;
}	/* SigUsr1*/




static void
SigAlrm(int sigNum)
{
	gGotSig = sigNum;
	longjmp(gJmp, 1);
}	/* SigAlrm */




static void
Out(char *buf, size_t len)
{
	unsigned char *lim;
	unsigned char *cp;

	if (gPrintableOnly == '.') {
		for (cp = (unsigned char *) buf, lim = cp + len; cp < lim; cp++) {
			if (!isprint(*cp) && (!isspace(*cp))) {
				*cp = '.';
			}
		}
		write(1, buf, len);
	} else if (gPrintableOnly == 'x') {
		for (cp = (unsigned char *) buf, lim = cp + len; cp < lim; cp++) {
			if (isprint(*cp) || (isspace(*cp))) {
				putchar(*cp);
			} else {
				fprintf(stdout, "$%X", ((unsigned int) *cp) & 0xff);
			}
		}
		fflush(stdout);
	} else if (gPrintableOnly == 'o') {
		for (cp = (unsigned char *) buf, lim = cp + len; cp < lim; cp++) {
			if (isprint(*cp) || (isspace(*cp))) {
				putchar(*cp);
			} else if (*cp == '\\') {
				putchar(*cp);
				putchar(*cp);
			} else {
				fprintf(stdout, "\\%o", ((unsigned int) *cp) & 0xff);
			}
		}
		fflush(stdout);
	} else {
		write(1, buf, len);
	}
}	/* Out */




static void
Client(void)
{
	int connfd;
	int nread;
	int nwrote;
	volatile int tread, twrote;
	char buf[512];
	struct sockaddr_in cliAddr;

	connfd = socket(gServerAddr.sin_family, SOCK_STREAM, 0);
	if (connfd < 0) {
		Perror("stream socket");
		exit(1);
	}

	cliAddr.sin_family = AF_INET;
	cliAddr.sin_addr.s_addr = INADDR_ANY;
	cliAddr.sin_port = htons((short) gOriginatingPort);
	if ((gOriginatingPort > 0) && (bind(connfd, (struct sockaddr *) &cliAddr, sizeof(cliAddr)) != 0)) {
		Perror("bind");
	}

	if (connect(connfd, (struct sockaddr *) &gServerAddr, sizeof(gServerAddr)) < 0) {
		Perror("connect");
		exit(1);
	}

	if (gPause != 0) {
		fprintf(stderr, "(paused; send SIGUSR1 to pid %d to continue)\n", (int) getpid());
		pause();
	}
	fprintf(stderr, "(connected; writing stdin to remote server)\n");
	for (twrote = 0; ; twrote += nwrote) {
		nread = read(0, buf, sizeof(buf));
		if (nread < 0) {
			Perror("read");
			exit(1);
		}
		if (nread == 0)
			break;
		nwrote = write(connfd, buf, nread);
		if (nwrote < 0) {
			Perror("write");
			exit(1);
		}
		if (nwrote == 0) {
			Perror("write(eof)");
			exit(1);
		}
	}

	signal(SIGALRM, SigAlrm);
	fprintf(stderr, "(reading reply back from server)\n");
	for (tread = 0; ; tread += nread) {
		if (setjmp(gJmp) != 0) {
			fprintf(stderr, "\n(timeout of %d seconds reached; no more data to read from server?)\n", gReadTimeout);
			break;
		}
		alarm(gReadTimeout);
		nread = read(connfd, buf, sizeof(buf));
		alarm(0);
		if (nread < 0) {
			Perror("read");
			exit(1);
		}
		if (nread == 0)
			break;
		Out(buf, (size_t) nread);
	}
	close(connfd);
	if (isatty(1))
		putchar('\n');
	fprintf(stderr, "(done; total written to server = %d, total read from server = %d)\n", twrote, tread);
}	/* Client */




static void
Server(short port)
{
	int servfd, connfd;
	int nread;
	int nwrote;
	volatile int tread, twrote;
	int caSize;
	struct sockaddr_in clientAddr;
	char buf[512];
	int i, on;
	int onsize;
	const char *volatile nl;

	gServerAddr.sin_family = AF_INET;
	gServerAddr.sin_addr.s_addr = htonl(INADDR_ANY);
	gServerAddr.sin_port = htons(port);

	servfd = socket(gServerAddr.sin_family, SOCK_STREAM, 0);
	if (servfd < 0) {
		Perror("datagram socket");
		exit(1);
	}

	/* This is mostly so you can quit the server and re-run it
	 * again right away.  If you don't do this, the OS may complain
	 * that the address is still in use.
	 */
	on = 1;
	onsize = (int) sizeof(on);
	(void) setsockopt(servfd, SOL_SOCKET, SO_REUSEADDR,
		(char *) &on, onsize);

	for (i=1; ; i++) {
		/* Try binding a few times, in case we get Address in Use
		 * errors.
		 */
		if (bind(servfd, (struct sockaddr *) &gServerAddr, sizeof(gServerAddr)) == 0) {
			break;
		}
		if (i == 3) {
			Perror("Can't bind local address, exiting");
			exit(1);
		} else {
			Perror("Can't bind local address, waiting 7 seconds");
			/* Give the OS time to clean up the old socket,
			 * and then try again.
			 */
			sleep(7);
		}
	}

	if (listen(servfd, 1) < 0) {
		Perror("listen");
		exit(1);
	}

	fprintf(stderr, "(waiting for connection from remote client)\n");
	caSize = (int) sizeof(clientAddr);
	connfd = accept(servfd, (struct sockaddr *) &clientAddr, &caSize);
	if (connfd < 0) {
		Perror("can't accept new connection");
		exit(1);
	}
	close(servfd);

	signal(SIGALRM, SigAlrm);
	fprintf(stderr, "(connected; reading request from client)\n");
	nl = "\n";
	for (tread = 0; ; tread += nread) {
		if (setjmp(gJmp) != 0) {
			fprintf(stderr, "\n(timeout of %d seconds reached; no more data to read from client?)\n", gReadTimeout);
			nl = "";
			break;
		}
		alarm(gReadTimeout);
		nread = read(connfd, buf, sizeof(buf));
		alarm(0);
		if (nread < 0) {
			Perror("read");
			exit(1);
		}
		if (nread == 0)
			break;
		Out(buf, (size_t) nread);
	}

	if (gPause != 0) {
		fprintf(stderr, "%s(paused; send SIGUSR1 to pid %d to continue)\n", nl, (int) getpid());
		pause();
		nl = "";
	}
	fprintf(stderr, "%s(sending stdin as reply to client)\n", nl);
	for (twrote = 0; ; twrote += nwrote) {
		nread = read(0, buf, sizeof(buf));
		if (nread < 0) {
			Perror("read");
			exit(1);
		}
		if (nread == 0)
			break;
		nwrote = write(connfd, buf, nread);
		if (nwrote < 0) {
			Perror("write");
			exit(1);
		}
		if (nwrote == 0) {
			Perror("write(eof)");
			exit(1);
		}
	}
	close(connfd);
	if (isatty(1))
		putchar('\n');
	fprintf(stderr, "(done; total read from client = %d, total written to client = %d)\n", tread, twrote);
}	/* Server */



static int
StrAddr(char *s, struct sockaddr_in *sa, int defaultport)
{
	char portstr[64];
	unsigned long ipnum;
	unsigned int port;
	struct hostent *hp;
	char *hostcp, *atsign, *colon;

	strncpy(portstr, s, sizeof(portstr));
	portstr[sizeof(portstr) - 1] = '\0';

	atsign = strchr(portstr, '@');
	if (atsign != NULL) {
		/* Look for port@host.name.domain */
		*atsign = '\0';
		hostcp = atsign + 1;
		port = (unsigned int) atoi(portstr);
	} else if ((colon = strchr(portstr, ':')) != NULL) {
		/* Look for host.name.domain:port */
		*colon = '\0';
		hostcp = portstr;
		port = (unsigned int) atoi(colon + 1);
	} else if (defaultport > 0) {
		/* Have just host.name.domain, use that w/ default port. */
		port = (unsigned int) defaultport;
		hostcp = portstr;
	} else {
		/* If defaultport <= 0, they must supply a port number
		 * in the host/port string.
		 */
		return (-1);
	}

	memset(sa, 0, sizeof(struct sockaddr_in));
	sa->sin_port = htons((short) port);

	ipnum = inet_addr(hostcp);
	if (ipnum != INADDR_NONE) {
		sa->sin_family = AF_INET;
		sa->sin_addr.s_addr = ipnum;
	} else {
		hp = gethostbyname(hostcp);
		if (hp == NULL) {
			fprintf(stderr, "Unknown host: %s", hostcp);
			return (-1);
		}
		sa->sin_family = hp->h_addrtype;
		memcpy(&sa->sin_addr.s_addr, hp->h_addr_list[0],
			(size_t) hp->h_length);
	}
	return (0);
}	/* StrAddr */




static void
Usage(void)
{
	fprintf(stderr, "Usages:  %s -s port [flags]\n", gProgName);
	fprintf(stderr, "         %s -c dest.host.name:port [flags]\n", gProgName);
	fprintf(stderr, "Flags:\n");
	fprintf(stderr, "  -p    : Substitute non-printing chars with \".\"\n");
	fprintf(stderr, "  -x    : Substitute non-printing chars with a hex representation\n");
	fprintf(stderr, "  -o    : Substitute non-printing chars with an octal representation\n");
	fprintf(stderr, "  -t NN : Set the read-timeout to NN seconds.\n");
	fprintf(stderr, "  -w    : Pause before writing to remote.\n");
	fprintf(stderr, "  -O NN : Bind local endpoint of client connection to port NN.\n");
	exit(2);
}	/* Usage */



int
main(int argc, char **argv)
{
	int i, c, type;
	short port = 0;

	gProgName = argv[0];
	type = '?';
	gPrintableOnly = isatty(1) ? '.' : 0;

	while ((c = getopt(argc, argv, "s:c:t:wxopO:l:")) > 0) switch(c) {
		case 'c':
			type = c;
			if (StrAddr(optarg, &gServerAddr, 0) < 0) {
				fprintf(stderr, "Bad host address: [%s]\n", optarg);
				Usage();
			}
			break;
		case 's':
			type = c;
			port = (short) atoi(optarg);
			if (port < 1)
				Usage();
			break;
		case 't':
			i = atoi(optarg);
			if (i < 0)		/* To block forever, use zero. */
				gReadTimeout = 0;
			else
				gReadTimeout = (unsigned int) i;
			break;
		case 'x':
		case 'o':
			gPrintableOnly = c;
			break;
		case 'p':
			gPrintableOnly = '.';
			break;
		case 'w':
			gPause = 1;
			break;
		case 'l':
		case 'O':
			gOriginatingPort = atoi(optarg);
			if (gOriginatingPort < 1)
				Usage();
			break;
		default:
			Usage();
	}

	signal(SIGUSR1, SigUsr1);
	signal(SIGUSR2, SigUsr1);
	if (type == 's')
		Server(port);
	else if (type == 'c')
		Client();
	else
		Usage();
	exit(0);
}	/* main */
