/* telnets.c
 *
 * Copyright (C) 1997 by Mike Gleason.
 * All Rights Reserved.
 *
 * This may be used and distributed freely.
 *
 */


#ifdef HAVE_CONFIG_H
#	include "config.h"
#else
#	define HAVE_STRERROR 1
#	define HAVE_UNISTD_H 1
#endif

#if defined(AIX) || defined(_AIX)
#	define _ALL_SOURCE 1
#endif

#ifdef HAVE_UNISTD_H
#	include <unistd.h>
#endif
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <errno.h>
#include <fcntl.h>
#ifdef CAN_USE_SYS_SELECT_H
#	include <sys/select.h>
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <signal.h>
#include <setjmp.h>
#include <time.h>
#include <ctype.h>

struct sockaddr_in gServerAddr;
char *gProgName;
jmp_buf gJmp;
int gPrintableOnly = 0;

extern int optind;
extern char *optarg;

static void
Perror(const char *const err)
{
	if (errno <= 0)
		fprintf(stderr, "%s :: %s.\n", gProgName, err);
	else {
#ifdef HAVE_STRERROR
		fprintf(stderr, "%s :: %s : %s.\n",
			gProgName, err, strerror(errno));
#else
		fprintf(stderr, "%s :: %s (errno=%d).\n",
			gProgName, err, errno);
#endif
	}
}	/* Perror */



static int
AddrToAddrStr(char *dst, size_t dsize, struct sockaddr_in * const saddrp, int dns)
{
	const char *addrNamePtr;
	struct hostent *hp;
	char str[128];

	if (dns == 0) {
		addrNamePtr = inet_ntoa(saddrp->sin_addr);
	} else {
		hp = gethostbyaddr((char *) &saddrp->sin_addr, (int) sizeof(struct in_addr), AF_INET);
		if (hp != NULL) {
			addrNamePtr = hp->h_name;
		} else {
			addrNamePtr = inet_ntoa(saddrp->sin_addr);
		}
	}
	if (addrNamePtr != NULL) {
		sprintf(str, "%d@%.120s",
			ntohs(saddrp->sin_port),
			addrNamePtr
		);
		strncpy(dst, str, dsize - 1);
		dst[dsize - 1] = '\0';
	} else {
		errno = EADDRNOTAVAIL;
		*dst = '\0';
		return (-1);
	}
	return (0);
}	/* AddrToAddrStr */




static void
Out(char *buf, size_t len)
{
	unsigned char *lim;
	unsigned char *cp;

	if (gPrintableOnly == '.') {
		for (cp = (unsigned char *) buf, lim = cp + len; cp < lim; cp++) {
			if (!isprint(*cp) && (!isspace(*cp))) {
				*cp = '.';
			}
		}
		write(1, buf, len);
	} else if (gPrintableOnly == 'x') {
		for (cp = (unsigned char *) buf, lim = cp + len; cp < lim; cp++) {
			if (isprint(*cp) || (isspace(*cp))) {
				putchar(*cp);
			} else {
				fprintf(stdout, "$%X", ((unsigned int) *cp) & 0xff);
			}
		}
		fflush(stdout);
	} else if (gPrintableOnly == 'o') {
		for (cp = (unsigned char *) buf, lim = cp + len; cp < lim; cp++) {
			if (isprint(*cp) || (isspace(*cp))) {
				putchar(*cp);
			} else if (*cp == '\\') {
				putchar(*cp);
				putchar(*cp);
			} else {
				fprintf(stdout, "\\%o", ((unsigned int) *cp) & 0xff);
			}
		}
		fflush(stdout);
	} else {
		write(1, buf, len);
	}
}	/* Out */




static void
Server(short port)
{
	int servfd, connfd;
	int nread;
	int nwrote;
	int caSize;
	struct sockaddr_in clientAddr;
	char buf[512];
	char caddrStr[128];
	int i, on;
	int onsize;
	fd_set ss0, ss;

	gServerAddr.sin_family = AF_INET;
	gServerAddr.sin_addr.s_addr = htonl(INADDR_ANY);
	gServerAddr.sin_port = htons(port);

	servfd = socket(gServerAddr.sin_family, SOCK_STREAM, 0);
	if (servfd < 0) {
		Perror("datagram socket");
		exit(1);
	}

	/* This is mostly so you can quit the server and re-run it
	 * again right away.  If you don't do this, the OS may complain
	 * that the address is still in use.
	 */
	on = 1;
	onsize = (int) sizeof(on);
	(void) setsockopt(servfd, SOL_SOCKET, SO_REUSEADDR,
		(char *) &on, onsize);

	for (i=1; ; i++) {
		/* Try binding a few times, in case we get Address in Use
		 * errors.
		 */
		if (bind(servfd, (struct sockaddr *) &gServerAddr, sizeof(gServerAddr)) == 0) {
			break;
		}
		if (i == 3) {
			Perror("Can't bind local address, exiting");
			exit(1);
		} else {
			Perror("Can't bind local address, waiting 7 seconds");
			/* Give the OS time to clean up the old socket,
			 * and then try again.
			 */
			sleep(7);
		}
	}

	if (listen(servfd, 1) < 0) {
		Perror("listen");
		exit(1);
	}

	fprintf(stderr, "(waiting for connection from remote client)\n");
	caSize = (int) sizeof(clientAddr);
	connfd = accept(servfd, (struct sockaddr *) &clientAddr, &caSize);
	if (connfd < 0) {
		Perror("can't accept new connection");
		exit(1);
	}
	close(servfd);

	memset(buf, 0, sizeof(buf));

	if (AddrToAddrStr(caddrStr, sizeof(caddrStr), &clientAddr, 1) == 0) {
		fprintf(stderr, "(connected to %s)\n", caddrStr);
	} else {
		fprintf(stderr, "(connected!)\n");
	}

	FD_ZERO(&ss0);
	FD_SET(0, &ss0);
	FD_SET(connfd, &ss0);

	for (;;) {
		memcpy(&ss, &ss0, sizeof(ss));
		if (select(connfd + 1, &ss, NULL, NULL, NULL) != 1)
			continue;
		if (FD_ISSET(0, &ss)) {
			nread = read(0, buf, sizeof(buf) - 1);
			if (nread == 0) {
				fprintf(stderr, "(you have closed the connection)\n");
				exit(0);
			}
			nwrote = write(connfd, buf, nread);
		} else if (FD_ISSET(connfd, &ss)) {
			nread = read(connfd, buf, sizeof(buf) - 1);
			if (nread == 0) {
				fprintf(stderr, "(remote host has closed the connection)\n");
				exit(0);
			}
			Out(buf, (size_t) nread);
		}
	}
}	/* Server */




static void
Usage(void)
{
	fprintf(stderr, "Usages:  %s -s port [flags]\n", gProgName);
	fprintf(stderr, "Flags:\n");
	fprintf(stderr, "  -p    : Substitute non-printing chars with \".\"\n");
	fprintf(stderr, "  -x    : Substitute non-printing chars with a hex representation\n");
	fprintf(stderr, "  -o    : Substitute non-printing chars with an octal representation\n");
	fprintf(stderr, "  -w    : Pause before writing to remote.\n");
	exit(2);
}	/* Usage */




int
main(int argc, char **argv)
{
	int c, type;
	short port = 5000;

	gProgName = argv[0];
	type = '?';
	gPrintableOnly = isatty(1) ? '.' : 0;

	while ((c = getopt(argc, argv, "s:xop")) > 0) switch(c) {
		case 's':
			type = c;
			port = (short) atoi(optarg);
			if (port < 1)
				Usage();
			break;
		case 'x':
		case 'o':
			gPrintableOnly = c;
			break;
		case 'p':
			gPrintableOnly = '.';
			break;
		default:
			Usage();
	}

	if (type == 's')
		Server(port);
	else
		Usage();
	exit(0);
}	/* main */
