/* udpcat.c
 *
 * Copyright (C) 1996 by Mike Gleason.
 * All Rights Reserved.
 *
 * This may be used and distributed freely.
 *
 */


#ifdef HAVE_CONFIG_H
#	include "config.h"
#else
#	define HAVE_STRERROR 1
#endif

#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <errno.h>
#include <fcntl.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <signal.h>
#include <setjmp.h>
#include <time.h>
#include <ctype.h>

#ifndef INADDR_NONE
#	define INADDR_NONE ((unsigned long)(-1L))
#endif

struct sockaddr_in gServerAddr;
char *gProgName;
jmp_buf gJmp;
unsigned int gReadTimeout = 10;
int gPrintableOnly = 0;
int gPause = 0;
int gSuck = 0;
int gGotSig = 0;
char gDatagramBuf[64 * 1024];
size_t gDatagramBufLen = 0;

extern int optind;
extern char *optarg;

static void
Perror(const char *const err)
{
	if (errno <= 0)
		fprintf(stderr, "%s :: %s.\n", gProgName, err);
	else {
#ifdef HAVE_STRERROR
		fprintf(stderr, "%s :: %s : %s.\n",
			gProgName, err, strerror(errno));
#else
		fprintf(stderr, "%s :: %s (errno=%d).\n",
			gProgName, err, errno);
#endif
	}
}	/* Perror */




static void
SigUsr1(int sigNum)
{
	gGotSig = sigNum;
	return;
}	/* SigUsr1*/




static void
SigAlrm(int sigNum)
{
	gGotSig = sigNum;
	longjmp(gJmp, 1);
}	/* SigAlrm */




static void
Out(char *buf, size_t len)
{
	unsigned char *lim;
	unsigned char *cp;

	if (gPrintableOnly == '.') {
		for (cp = (unsigned char *) buf, lim = cp + len; cp < lim; cp++) {
			if (!isprint(*cp) && (!isspace(*cp))) {
				*cp = '.';
			}
		}
		write(1, buf, len);
	} else if (gPrintableOnly == 'x') {
		for (cp = (unsigned char *) buf, lim = cp + len; cp < lim; cp++) {
			if (isprint(*cp) || (isspace(*cp))) {
				putchar(*cp);
			} else {
				fprintf(stdout, "$%X", ((unsigned int) *cp) & 0xff);
			}
		}
		fflush(stdout);
	} else if (gPrintableOnly == 'o') {
		for (cp = (unsigned char *) buf, lim = cp + len; cp < lim; cp++) {
			if (isprint(*cp) || (isspace(*cp))) {
				putchar(*cp);
			} else if (*cp == '\\') {
				putchar(*cp);
				putchar(*cp);
			} else {
				fprintf(stdout, "\\%o", ((unsigned int) *cp) & 0xff);
			}
		}
		fflush(stdout);
	} else {
		write(1, buf, len);
	}
}	/* Out */




static void
StdinToBuf(void)
{
	char *bufp;
	char *buflim;
	char subbuf[512];
	int nread;
	size_t ntocopy;

	bufp = gDatagramBuf;
	buflim = bufp + sizeof(gDatagramBuf);
	*bufp = '\0';
	while (bufp < buflim) {
		nread = read(0, subbuf, sizeof(subbuf));
		if (nread == 0)
			break;
		if (nread < 0) {
			if (errno == EINTR)
				continue;
			Perror("read");
			exit(1);
		}
		ntocopy = (size_t) nread;
		if (bufp + ntocopy > buflim)
			ntocopy = (size_t) (buflim - bufp);
		memcpy(bufp, subbuf, ntocopy);
		bufp += ntocopy;
	}
	gDatagramBufLen = (size_t) (bufp - gDatagramBuf);
}	/* StdinToBuf */




static void
Client(void)
{
	int connfd;
	int nread;
	int nwrote;
	struct sockaddr_in replyAddr;
	int replyAddrSize;

	connfd = socket(gServerAddr.sin_family, SOCK_DGRAM, 0);
	if (connfd < 0) {
		Perror("dgram socket");
		exit(1);
	}

	fprintf(stderr, "(writing stdin to remote server)\n");
	StdinToBuf();
	nwrote = sendto(connfd, gDatagramBuf, gDatagramBufLen, 0, (struct sockaddr *) &gServerAddr, sizeof(gServerAddr));
	if (nwrote < 0) {
		Perror("sendto");
		exit(1);
	}

	signal(SIGALRM, SigAlrm);
	fprintf(stderr, "(reading reply back from server)\n");

	if (setjmp(gJmp) != 0) {
		fprintf(stderr, "\n(timeout of %d seconds reached; no data to read from server?)\n", gReadTimeout);
		nread = 0;
	} else {
		replyAddrSize = sizeof(replyAddr);
		alarm(gReadTimeout);
		nread = recvfrom(connfd, gDatagramBuf, sizeof(gDatagramBuf), 0, (struct sockaddr *) &replyAddr, &replyAddrSize);
	}
	alarm(0);
	if (nread < 0) {
		Perror("recvfrom");
		exit(1);
	}
	if (nread > 0)
		Out(gDatagramBuf, (size_t) nread);
	close(connfd);
	if (isatty(1))
		putchar('\n');
	fprintf(stderr, "(done; total written to server = %d, total read from server = %d)\n", nwrote, nread);
}	/* Client */




static void
Server(short port)
{
	struct sockaddr_in replyAddr;
	int replyAddrSize;
	int servfd, connfd;
	int nread;
	int nwrote;
	int i, on;
	int onsize;
	const char *nl;

	gServerAddr.sin_family = AF_INET;
	gServerAddr.sin_addr.s_addr = htonl(INADDR_ANY);
	gServerAddr.sin_port = htons(port);

	servfd = socket(gServerAddr.sin_family, SOCK_DGRAM, 0);
	if (servfd < 0) {
		Perror("datagram socket");
		exit(1);
	}

	/* This is mostly so you can quit the server and re-run it
	 * again right away.  If you don't do this, the OS may complain
	 * that the address is still in use.
	 */
	on = 1;
	onsize = (int) sizeof(on);
	(void) setsockopt(servfd, SOL_SOCKET, SO_REUSEADDR,
		(char *) &on, onsize);

	for (i=1; ; i++) {
		/* Try binding a few times, in case we get Address in Use
		 * errors.
		 */
		if (bind(servfd, (struct sockaddr *) &gServerAddr, sizeof(gServerAddr)) == 0) {
			break;
		}
		if (i == 3) {
			Perror("Can't bind local address, exiting");
			exit(1);
		} else {
			Perror("Can't bind local address, waiting 7 seconds");
			/* Give the OS time to clean up the old socket,
			 * and then try again.
			 */
			sleep(7);
		}
	}

	do {
		fprintf(stderr, "(waiting for message from remote client)\n");
		connfd = servfd;

		signal(SIGALRM, SigAlrm);
		nl = "\n";

		replyAddrSize = sizeof(replyAddr);
		nread = recvfrom(connfd, gDatagramBuf, sizeof(gDatagramBuf), 0, (struct sockaddr *) &replyAddr, &replyAddrSize);
		if (nread < 0) {
			Perror("recvfrom");
			exit(1);
		}
		if (nread > 0)
			Out(gDatagramBuf, (size_t) nread);

		if (!gSuck) {
			if (gPause != 0) {
				fprintf(stderr, "%s(paused; send SIGUSR1 to pid %d to continue)\n", nl, (int) getpid());
				pause();
				nl = "";
			}

			fprintf(stderr, "%s(sending stdin as reply to client)\n", nl);
			StdinToBuf();
			nwrote = sendto(connfd, gDatagramBuf, gDatagramBufLen, 0, (struct sockaddr *) &replyAddr, sizeof(replyAddr));
		}
	} while (gSuck != 0);

	close(connfd);
	if (isatty(1))
		putchar('\n');
	fprintf(stderr, "(done; total read from client = %d, total written to client = %d)\n", nread, nwrote);
}	/* Server */



static int
StrAddr(char *s, struct sockaddr_in *sa, int defaultport)
{
	char portstr[64];
	unsigned long ipnum;
	unsigned int port;
	struct hostent *hp;
	char *hostcp, *atsign, *colon;

	strncpy(portstr, s, sizeof(portstr));
	portstr[sizeof(portstr) - 1] = '\0';

	atsign = strchr(portstr, '@');
	if (atsign != NULL) {
		/* Look for port@host.name.domain */
		*atsign = '\0';
		hostcp = atsign + 1;
		port = (unsigned int) atoi(portstr);
	} else if ((colon = strchr(portstr, ':')) != NULL) {
		/* Look for host.name.domain:port */
		*colon = '\0';
		hostcp = portstr;
		port = (unsigned int) atoi(colon + 1);
	} else if (defaultport > 0) {
		/* Have just host.name.domain, use that w/ default port. */
		port = (unsigned int) defaultport;
		hostcp = portstr;
	} else {
		/* If defaultport <= 0, they must supply a port number
		 * in the host/port string.
		 */
		return (-1);
	}

	memset(sa, 0, sizeof(struct sockaddr_in));
	sa->sin_port = htons((short) port);

	ipnum = inet_addr(hostcp);
	if (ipnum != INADDR_NONE) {
		sa->sin_family = AF_INET;
		sa->sin_addr.s_addr = ipnum;
	} else {
		hp = gethostbyname(hostcp);
		if (hp == NULL) {
			fprintf(stderr, "Unknown host: %s", hostcp);
			return (-1);
		}
		sa->sin_family = hp->h_addrtype;
		memcpy(&sa->sin_addr.s_addr, hp->h_addr_list[0],
			(size_t) hp->h_length);
	}
	return (0);
}	/* StrAddr */




static void
Usage(void)
{
	fprintf(stderr, "Usages:  %s -s port [flags]\n", gProgName);
	fprintf(stderr, "         %s -c dest.host.name:port [flags]\n", gProgName);
	fprintf(stderr, "Flags:\n");
	fprintf(stderr, "  -p    : Substitute non-printing chars with \".\"\n");
	fprintf(stderr, "  -x    : Substitute non-printing chars with a hex representation\n");
	fprintf(stderr, "  -o    : Substitute non-printing chars with an octal representation\n");
	fprintf(stderr, "  -t NN : Set the read-timeout to NN seconds.\n");
	fprintf(stderr, "  -w    : Pause before writing to remote.\n");
	fprintf(stderr, "  -S    : Suck packets continuously.\n");
	exit(2);
}	/* Usage */



int
main(int argc, char **argv)
{
	int c, type;
	short port = 5000;

	gProgName = argv[0];
	type = '?';
	gPrintableOnly = isatty(1) ? '.' : 0;

	while ((c = getopt(argc, argv, "s:c:t:wxopS")) > 0) switch(c) {
		case 'c':
			type = c;
			if (StrAddr(optarg, &gServerAddr, 0) < 0) {
				fprintf(stderr, "Bad host address: [%s]\n", optarg);
				Usage();
			}
			break;
		case 's':
			type = c;
			port = (short) atoi(optarg);
			if (port < 1)
				Usage();
			break;
		case 't':
			c = atoi(optarg);
			if (c < 0)		/* To block forever, use zero. */
				gReadTimeout = 0;
			else
				gReadTimeout = c;
			break;
		case 'x':
		case 'o':
			gPrintableOnly = c;
			break;
		case 'p':
			gPrintableOnly = '.';
			break;
		case 'w':
			gPause = 1;
			break;
		case 'S':
			gSuck = 1;
			break;
		default:
			Usage();
	}

	signal(SIGUSR1, SigUsr1);
	signal(SIGUSR2, SigUsr1);
	if (type == 's')
		Server(port);
	else if (type == 'c')
		Client();
	else
		Usage();
	exit(0);
}	/* main */
